"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
const node_gyp_build_1 = __importDefault(require("node-gyp-build"));
const package_json_1 = require("../package.json");
const { LockRequest, snapshot } = node_gyp_build_1.default('.');
const GRANTED = 0;
const NOT_AVAILABLE = 1;
const CANCELED = 2;
function onabort(abortSignal, listener) {
    if ('addEventListener' in abortSignal) {
        abortSignal.addEventListener('abort', listener, { once: true });
    }
    else {
        abortSignal.once('abort', listener);
    }
}
function clearAbort(abortSignal, listener) {
    if ('addEventListener' in abortSignal) {
        abortSignal.removeEventListener('abort', listener, { once: true });
    }
    else {
        abortSignal.off('abort', listener);
    }
}
class AbortError extends Error {
    constructor() {
        super('The task has been aborted');
    }
    get name() { return 'AbortError'; }
}
const kDefaultRequestOptions = {
    mode: 'exclusive',
    ifAvailable: false,
    steal: false,
    signal: undefined
};
async function request(name, options, callback) {
    if (typeof options === 'function') {
        callback = options;
        options = kDefaultRequestOptions;
    }
    name = `${name}`;
    if (typeof callback !== 'function') {
        throw new TypeError('Callback must be a function');
    }
    if (options == null || typeof options !== 'object') {
        throw new TypeError('Options must be an object');
    }
    options = { ...kDefaultRequestOptions, ...options };
    let request;
    const abortListener = () => request.cancel();
    if (options.mode !== 'exclusive' && options.mode !== 'shared') {
        throw new RangeError('Invalid mode');
    }
    if (typeof options.ifAvailable !== 'boolean') {
        throw new TypeError('options.ifAvailable must be a boolean');
    }
    if (typeof options.steal !== 'boolean') {
        throw new TypeError('options.steal must be a boolean');
    }
    if (options.signal != null) {
        if (options.signal.aborted) {
            return Promise.reject(new AbortError());
        }
        onabort(options.signal, abortListener);
    }
    const lock = await new Promise((resolve, reject) => {
        request = new LockRequest(name, options.mode === 'exclusive' ? 0 : 1, Boolean(options.ifAvailable), Boolean(options.steal), (status, lock) => {
            if (options.signal != null) {
                clearAbort(options.signal, abortListener);
            }
            request = undefined;
            switch (status) {
                case GRANTED: return resolve(lock);
                case NOT_AVAILABLE: return resolve(null);
                case CANCELED: return reject(new AbortError());
            }
        });
    });
    try {
        return await callback(lock);
    }
    finally {
        if (lock != null) {
            lock.release();
        }
    }
}
function query() {
    return snapshot();
}
module.exports = {
    request,
    query,
    version: package_json_1.version
};
//# sourceMappingURL=index.js.map